// Copyright 2011 Boris Kogan (boris@thekogans.net)
//
// This file is part of libthekogans_util.
//
// libthekogans_util is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// libthekogans_util is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libthekogans_util. If not, see <http://www.gnu.org/licenses/>.

#if !defined (__thekogans_util_Console_h)
#define __thekogans_util_Console_h

#include <string>
#include "thekogans/util/Config.h"
#include "thekogans/util/Singleton.h"
#include "thekogans/util/SpinLock.h"
#include "thekogans/util/Event.h"
#include "thekogans/util/JobQueue.h"
#include "thekogans/util/TimeSpec.h"

namespace thekogans {
    namespace util {

        /// \struct Console Console.h thekogans/util/Console.h
        ///
        /// \brief
        /// Console provides platform independent CTRL+BREAK handling and
        /// colored text output. On Linux and OS X it also turns on core
        /// dumping, and ignores the SIGPIPE.

        struct _LIB_THEKOGANS_UTIL_DECL Console :
                public Singleton<Console, SpinLock> {
        private:
            /// \vrief
            /// If true, use jobQueue to serialize print requests.
            bool threadSafePrintString;
            /// \brief
            /// Event to signal when CTRL+BREAK was signalled.
            Event ctrlBreak;
            /// \brief
            /// Used to synchronize access to cout and cerr in PrintString.
            JobQueue jobQueue;

        public:
            /// \brief
            /// ctor.
            Console ();

            /// \brief
            /// Set the desired print thread safety.
            /// \param[in] threadSafePrintString_ true = Serialize print requests.
            inline void SetThreadSafePrintString (bool threadSafePrintString_) {
                threadSafePrintString = threadSafePrintString_;
            }

            /// \brief
            /// Wait until ctrlBreak event is signaled.
            void WaitCtrlBreak ();

            /// \brief
            /// Wait until ctrlBreak event is signaled.
            /// \param[in] timeSpec How long to wait.
            /// \return true = ctrlBreak was signalled, false = timed out.
            bool WaitCtrlBreak (const TimeSpec &timeSpec);

            /// \brief
            /// Signal ctrlBreak.
            void CtrlBreak ();

            enum {
                /// \brief
                /// Print to cout.
                PRINT_COUT,
                /// \brief
                /// Print to cerr.
                PRINT_CERR
            };

        #if defined (TOOLCHAIN_OS_Windows)
            /// \brief
            /// Color type.
            typedef WORD ColorType;
        #else // defined (TOOLCHAIN_OS_Windows)
            /// \brief
            /// Color type.
            typedef char *ColorType;
        #endif // defined (TOOLCHAIN_OS_Windows)
            /// \brief
            /// Red.
            static const ColorType TEXT_COLOR_RED;
            /// \brief
            /// Greeen.
            static const ColorType TEXT_COLOR_GREEN;
            /// \brief
            /// Yellow.
            static const ColorType TEXT_COLOR_YELLOW;
            /// \brief
            /// Blue.
            static const ColorType TEXT_COLOR_BLUE;
            /// \brief
            /// Magenta.
            static const ColorType TEXT_COLOR_MAGENTA;
            /// \brief
            /// Cyan.
            static const ColorType TEXT_COLOR_CYAN;
            /// \brief
            /// White.
            static const ColorType TEXT_COLOR_WHITE;

            /// \brief
            /// Print a string to cout or cerr.
            /// \param[in] str String to print.
            /// \param[in] where PRINT_COUT or PRINT_CERR.
            /// \param[in] color Text color.
            void PrintString (
                const std::string &str,
                ui32 where = PRINT_COUT,
                const ColorType color = 0);

            /// \brief
            /// Console is neither copy constructable, nor assignable.
            THEKOGANS_UTIL_DISALLOW_COPY_AND_ASSIGN (Console)
        };

    } // namespace util
} // namespace thekogans

#endif // !defined (__thekogans_util_Console_h)
