// Copyright 2011 Boris Kogan (boris@thekogans.net)
//
// This file is part of libthekogans_util.
//
// libthekogans_util is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// libthekogans_util is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libthekogans_util. If not, see <http://www.gnu.org/licenses/>.

#if !defined (__thekogans_util_File_h)
#define __thekogans_util_File_h

#if defined (TOOLCHAIN_OS_Windows) 
    #if !defined (_WINDOWS_)
        #if !defined (WIN32_LEAN_AND_MEAN)
            #define WIN32_LEAN_AND_MEAN
        #endif // !defined (WIN32_LEAN_AND_MEAN)
        #if !defined (NOMINMAX)
            #define NOMINMAX
        #endif // !defined (NOMINMAX)
        #include <windows.h>
    #endif // !defined (_WINDOWS_)
#else // defined (TOOLCHAIN_OS_Windows)
    #if defined (TOOLCHAIN_OS_Linux)
        #if !defined (_LARGEFILE64_SOURCE)
            #define _LARGEFILE64_SOURCE 1
        #endif // !defined (_LARGEFILE64_SOURCE)
        #define ftruncate ftruncate64
        #define lseek lseek64
    #endif // defined (TOOLCHAIN_OS_Linux)
    #include <fcntl.h>
    #include <sys/types.h>
    #include <sys/time.h>
    #include <unistd.h>
#endif // defined (TOOLCHAIN_OS_Windows)
#include <cstdio>
#include <string>
#include "thekogans/util/Config.h"
#include "thekogans/util/Types.h"
#include "thekogans/util/Serializer.h"
#include "thekogans/util/Buffer.h"
#include "thekogans/util/GUID.h"

namespace thekogans {
    namespace util {

        /// \struct File File.h thekogans/util/File.h
        ///
        /// \brief
        /// File is a platform independent file system file.

        struct _LIB_THEKOGANS_UTIL_DECL File : public Serializer {
        protected:
            /// \brief
            /// OS specific file handle.
            THEKOGANS_UTIL_HANDLE handle;
            /// \brief
            /// Path that was used to open the file.
            std::string path;

        public:
            /// \brief
            /// Default ctor.
            /// \param[in] endianness File endianness.
            /// \param[in] handle_ OS file handle.
            /// \param[in] path_ File path.
            File (
                Endianness endianness = HostEndian,
                THEKOGANS_UTIL_HANDLE handle_ = THEKOGANS_UTIL_INVALID_HANDLE_VALUE,
                const std::string &path_ = std::string ()) :
                Serializer (endianness),
                handle (handle_),
                path (path_) {}
        #if defined (TOOLCHAIN_OS_Windows)
            /// \brief ctor
            /// Open the file.
            /// \param[in] endianness File endianness.
            /// \param[in] path Path to file to open.
            /// \param[in] dwDesiredAccess Windows CreateFile parameter.
            /// \param[in] dwShareMode Windows CreateFile parameter.
            /// \param[in] dwCreationDisposition Windows CreateFile parameter.
            /// \param[in] dwFlagsAndAttributes Windows CreateFile parameter.
            File (
                Endianness endianness,
                const std::string &path,
                DWORD dwDesiredAccess = GENERIC_READ | GENERIC_WRITE,
                DWORD dwShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE,
                DWORD dwCreationDisposition = OPEN_ALWAYS,
                DWORD dwFlagsAndAttributes = FILE_ATTRIBUTE_NORMAL);
        #else // defined (TOOLCHAIN_OS_Windows)
            /// \brief ctor
            /// Open the file.
            /// \param[in] endianness File endianness.
            /// \param[in] path Path to file to open.
            /// \param[in] flags POSIX open parameter.
            /// \param[in] mode POSIX open parameter.
            File (
                Endianness endianness,
                const std::string &path,
                int flags = O_RDWR | O_CREAT,
                int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        #endif // defined (TOOLCHAIN_OS_Windows)
            /// \brief dtor
            /// Close the file.
            virtual ~File ();

        #if defined (TOOLCHAIN_OS_Windows)
            /// \brief
            /// Open the file.
            /// \param[in] path_ Windows CreateFile parameter.
            /// \param[in] dwDesiredAccess Windows CreateFile parameter.
            /// \param[in] dwShareMode Windows CreateFile parameter.
            /// \param[in] dwCreationDisposition Windows CreateFile parameter.
            /// \param[in] dwFlagsAndAttributes Windows CreateFile parameter.
            virtual void Open (
                const std::string &path_,
                DWORD dwDesiredAccess = GENERIC_READ | GENERIC_WRITE,
                DWORD dwShareMode = FILE_SHARE_READ | FILE_SHARE_WRITE,
                DWORD dwCreationDisposition = OPEN_ALWAYS,
                DWORD dwFlagsAndAttributes = FILE_ATTRIBUTE_NORMAL);
        #else // defined (TOOLCHAIN_OS_Windows)
            /// \brief
            /// Open the file.
            /// \param[in] path_ POSIX open parameter.
            /// \param[in] flags POSIX open parameter.
            /// \param[in] mode POSIX open parameter.
            virtual void Open (
                const std::string &path_,
                int flags = O_RDWR | O_CREAT,
                int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
        #endif // defined (TOOLCHAIN_OS_Windows)
            /// \brief
            /// Close file.
            virtual void Close ();
            /// \brief
            /// Flush pending writes to a file.
            virtual void Flush ();

            /// \brief
            /// Return number of bytes available for reading.
            /// \return Number of bytes available for reading.
            virtual ui64 GetDataAvailable ();

            // Serializer
            /// \brief
            /// Read bytes from a file.
            /// \param[out] buffer Where to place the bytes.
            /// \param[in] count Number of bytes to read.
            /// \return Number of bytes actually read.
            virtual ui32 Read (
                void *buffer,
                ui32 count);
            /// \brief
            /// Write bytes to a file.
            /// \param[in] buffer Where the bytes come from.
            /// \param[in] count Number of bytes to write.
            /// \return Number of bytes actually written.
            virtual ui32 Write (
                const void *buffer,
                ui32 count);

            /// \brief
            /// Return the file pointer position.
            /// \return The file pointer position.
            virtual i64 Tell () const;
            /// \brief
            /// Reposition the file pointer.
            /// \param[in] offset Offset to move relative to fromWhere.
            /// \param[in] fromWhere SEEK_SET, SEEK_CUR or SEEK_END.
            /// \return The new file pointer position.
            virtual i64 Seek (
                i64 offset,
                i32 fromWhere);
            /// \brief
            /// Return file size in bytes.
            /// \return File size in bytes.
            virtual ui64 GetSize () const;
            /// \brief
            /// Truncates or expands the file.
            /// \param[in] newSize New size to set the file to.
            virtual void SetSize (ui64 newSize);

            /// \struct File::Region File.h thekogans/util/File.h
            ///
            /// \brief
            /// Encapsulates file region lock/unlock parameters.
            struct Region {
                /// \brief
                /// File offset where the region starts.
                ui64 offset;
                /// \brief
                /// Region length.
                ui64 length;
                /// \brief
                /// ctor.
                /// \param[in] offset_ File offset where the region starts.
                /// \param[in] length_ Region length.
                Region (
                    ui64 offset_,
                    ui64 length_) :
                    offset (offset_),
                    length (length_) {}
            };
            /// \brief
            /// Lock a range of bytes in the file.
            /// \region[in] region region to lock.
            /// \region[in] exclusive lock for exclusive access.
            virtual void LockRegion (
                const Region &region,
                bool exclusive);
            /// \brief
            /// Unlock a range of bytes in the file.
            /// \region[in] region region to unlock.
            virtual void UnlockRegion (const Region &region);

            /// \brief
            /// Return the file path.
            /// \return File path.
            inline std::string GetPath () const {
                return path;
            }

            /// \brief
            /// Return a unique identifier for this file.
            /// \return Unique identifier for this file.
            GUID GetId () const;

            /// \brief
            /// Update the last modified time. If the given
            /// file does not exist, create it.
            /// \param[in] path File to touch.
            static void Touch (const std::string &path);

        private:
            /// \brief
            /// Platform specific seek.
            /// \param[in] offset Offset to move relative to fromWhere.
            /// \param[in] fromWhere SEEK_SET, SEEK_CUR or SEEK_END.
            /// \return The new file pointer position.
            i64 PlatformSeek (
                i64 offset,
                i32 fromWhere) const;

            /// \brief
            /// File is neither copy constructable, nor assignable.
            THEKOGANS_UTIL_DISALLOW_COPY_AND_ASSIGN (File)
        };

        /// \struct ReadOnlyFile File.h thekogans/util/File.h
        ///
        /// \brief
        /// As its name implies, ReadOnlyFile is a convenient
        /// platform independent read only file.

        struct ReadOnlyFile : public File {
            /// \brief
            /// ctor.
            /// \param[in] endianness File endianness.
            /// \param[in] path Path to file to open.
            explicit ReadOnlyFile (
                Endianness endianness,
                const std::string &path) :
            #if defined (TOOLCHAIN_OS_Windows)
                File (
                    endianness,
                    path,
                    GENERIC_READ,
                    FILE_SHARE_READ,
                    OPEN_EXISTING,
                    FILE_ATTRIBUTE_NORMAL) {}
            #else // defined (TOOLCHAIN_OS_Windows)
                File (
                    endianness,
                    path,
                    O_RDONLY) {}
            #endif // defined (TOOLCHAIN_OS_Windows)

            /// \brief
            /// ReadOnlyFile is neither copy constructable, nor assignable.
            THEKOGANS_UTIL_DISALLOW_COPY_AND_ASSIGN (ReadOnlyFile)
        };

        /// \struct TenantFile File.h thekogans/util/File.h
        ///
        /// \brief
        /// TenantFile wraps a raw file handle to provide the features
        /// of File without taking ownership.

        struct _LIB_THEKOGANS_UTIL_DECL TenantFile : public File {
            /// \brief
            /// ctor.
            /// \param[in] endianness File endianness.
            /// \param[in] handle File handle.
            /// \param[in] path File path.
            TenantFile (
                Endianness endianness,
                THEKOGANS_UTIL_HANDLE handle,
                const std::string &path) :
                File (endianness, handle, path) {}
            /// \brief
            /// dtor.
            /// Step on the handle so that ~File () don't close it.
            virtual ~TenantFile () {
                // Prevent File dtor from closing a descriptor
                // that does not belong to us.
                handle = THEKOGANS_UTIL_INVALID_HANDLE_VALUE;
            }

            /// \brief
            /// TenantFile is neither copy constructable, nor assignable.
            THEKOGANS_UTIL_DISALLOW_COPY_AND_ASSIGN (TenantFile)
        };

    } // namespace util
} // namespace thekogans

#endif // !defined (__thekogans_util_File_h)
