// Copyright 2011 Boris Kogan (boris@thekogans.net)
//
// This file is part of libthekogans_util.
//
// libthekogans_util is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// libthekogans_util is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libthekogans_util. If not, see <http://www.gnu.org/licenses/>.

#if !defined (__thekogans_util_MD5_h)
#define __thekogans_util_MD5_h

#include <string>
#include "thekogans/util/Config.h"
#include "thekogans/util/Types.h"
#include "thekogans/util/Hash.h"

namespace thekogans {
    namespace util {

        /// \struct MD5 MD5.h thekogans/util/MD5.h
        ///
        /// \brief
        /// Use instances of this class to create MD5 hashes.

        struct _LIB_THEKOGANS_UTIL_DECL MD5 : public Hash {
            /// \brief
            /// MD5 participates in the Hash dynamic discovery and creation.
            THEKOGANS_UTIL_DECLARE_HASH (MD5)

            enum {
                /// \brief
                /// MD5 hash size in bytes.
                DIGEST_SIZE_128 = 16
            };

        private:
            enum {
                /// \brief
                /// Size of the state vector in ui32.
                STATE_SIZE = 4,
                /// \brief
                /// Block size in bytes.
                BLOCK_SIZE = 64,
                /// \brief
                /// Block size, without the 8 bytes for bitCount, in bytes.
                SHORT_BLOCK_SIZE = BLOCK_SIZE - 8
            };
            /// \brief
            /// Incremental state used during hashing.
            ui32 state[STATE_SIZE];
            /// \brief
            /// Number of input bits processed.
            ui64 bitCount;
            /// \brief
            /// Current data being hashed.
            ui8 buffer[BLOCK_SIZE];
            /// \brief
            /// Index in to the buffer where next write will occur.
            ui32 bufferIndex;

        public:
            /// \brief
            /// ctor. Reset the hasher.
            MD5 () {
                Reset ();
            }

            /// \brief
            /// Return hasher name.
            /// \return Hasher name.
            virtual std::string GetName (ui32 /*digestSize*/) const {
                return "MD5";
            }
            /// \brief
            /// Return hasher supported digest sizes.
            /// \param[out] digestSizes List of supported digest sizes.
            virtual void GetDigestSizes (std::list<ui32> &digestSizes) const {
                digestSizes.push_back (DIGEST_SIZE_128);
            }
            /// \brief
            /// Initialize the hasher.
            virtual void Init (ui32 digestSize);
            /// \brief
            /// Hash a buffer. Call multiple times before
            /// Finalize to process incremental data.
            /// \param[in] buffer Buffer to hash.
            /// \param[in] size Size of buffer in bytes.
            virtual void Update (
                const void *buffer,
                std::size_t size);
            /// \brief
            /// Finalize the hashing operation, and retrieve the digest.
            /// \param[out] digest Result of the hashing operation.
            virtual void Final (Digest &digest);

        private:
            /// \brief
            /// Called internally after finalize to reset
            /// the hasher and prevent secret leaking.
            void Reset ();
            /// \brief
            /// Called internally to transform (hash) the input block.
            void Transform ();
        };

    } // namespace util
} // namespace thekogans

#endif // !defined (__thekogans_util_MD5_h)
