#include "CrownHandler.h"

#include <iostream>
#include <sstream>
#include <thread>

#include "json\json.h"

#include "..\input\InputManager.h"
#include <GLFW/glfw3.h>

using websocketpp::lib::placeholders::_1;
using websocketpp::lib::placeholders::_2;
using websocketpp::lib::bind;

using client = websocketpp::client<websocketpp::config::asio_client>;

const char *PluginGuid = "619ae85f-09c1-4a4e-b6ba-9b8d10d9fbf9";
const float Sensitivity = 0.25;

void onOpen(std::shared_ptr<client> c, websocketpp::connection_hdl hdl)
{
    std::cout << "onOpen called with hdl: " << hdl.lock().get() << std::endl;

    websocketpp::lib::error_code ec;

    // Send plugin registration message
    Json::Value json;
    json["message_type"] = "register";
    json["plugin_guid"] = PluginGuid;
    json["PID"] = (unsigned int)GetCurrentProcessId();
    json["execName"] = "breakout.exe";

    c->send(hdl, Json::FastWriter().write(json), websocketpp::frame::opcode::value::text, ec);
    if (ec)
    {
        std::cout << "Plugin registration failed because: " << ec.message() << std::endl;
    }
}

void onClose(websocketpp::connection_hdl hdl)
{
    std::cout << "onClose called with hdl: " << hdl.lock().get() << std::endl;
}

void onMessage(websocketpp::connection_hdl hdl, websocketpp::config::asio_client::message_type::ptr msg)
{
    std::cout << "onMessage called with hdl: " << hdl.lock().get() << " and message: " << msg->get_payload() << std::endl;

    Json::Value json;
    Json::Reader reader;

    if (!reader.parse(msg->get_payload(), json))
    {
        std::cout << "Error parsing message. Reason: " << reader.getFormatedErrorMessages();
        return;
    }

    if (json["message_type"].asString() == "crown_turn_event")
    {
        try
        {
            auto delta = json["delta"].asInt();
            InputManager::Instance().setAnalogMovement((float)delta * Sensitivity);

            auto ratchetDelta = json["ratchet_delta"].asInt();
            InputManager::Instance().setRatchetMovement(ratchetDelta);
        }
        catch (std::exception e)
        {
            std::cout << e.what() << std::endl;
        }
    }
}

void CrownHandler::startUp()
{
    std::string uri = "ws://localhost:10134";

    try
    {
        std::shared_ptr<client> c = std::make_shared<client>();

        // Set logging to be pretty verbose (everything except message payloads)
        c->set_access_channels(websocketpp::log::alevel::all);
        c->clear_access_channels(websocketpp::log::alevel::frame_payload);

        // Initialize ASIO
        c->init_asio();

        // Register our handlers
        c->set_message_handler(bind(&onMessage, ::_1, ::_2));
        c->set_open_handler(bind(&onOpen, c, ::_1));
        c->set_close_handler(bind(&onClose, ::_1));

        websocketpp::lib::error_code ec;
        client::connection_ptr con = c->get_connection(uri, ec);
        if (ec)
        {
            std::cout << "could not create connection because: " << ec.message() << std::endl;
            return;
        }

        // Note that connect here only requests a connection. No network messages are
        // exchanged until the event loop starts running in the next line.
        c->connect(con);

        // Start the ASIO io_service run loop
        // this will cause a single connection to be made to the server. 
        // c->run() will exit when this connection is closed.
        auto t = std::thread([c]()
        {
            // Thread this; don't block main thread
            c->run();
        });
        t.detach();
    }
    catch (websocketpp::exception const & e)
    {
        std::cout << e.what() << std::endl;
    }
}

void CrownHandler::shutDown()
{
    // TODO
}
